/*
Copyright (C) 2013 Bengt Martensson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see http://www.gnu.org/licenses/.
*/

package org.harctoolbox.irscrutinizer.importer;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.Reader;
import java.text.ParseException;
import java.util.Collection;
import org.harctoolbox.IrpMaster.IrSignal;
import org.harctoolbox.IrpMaster.IrpMasterException;
import org.harctoolbox.IrpMaster.IrpUtils;
import org.harctoolbox.girr.Command;
import org.harctoolbox.irscrutinizer.Utils;

/**
 * This class does something interesting and useful. Or not...
 *
 * Columns are numbered starting with 1.
 */
public class CsvRawImporter extends CsvImporter {
    private static final long serialVersionUID = 1L;
    private int codeColumn;
    private boolean includeTail;
    private boolean invokeRepeatFinder;

    /**
     * @param codeColumn the codeColumn to set
     */
    public void setCodeColumn(int codeColumn) {
        this.codeColumn = codeColumn;
    }

    /**
     * @param includeTail the includeTail to set
     */
    public void setIncludeTail(boolean includeTail) {
        this.includeTail = includeTail;
    }

    /**
     * @param invokeRepeatFinder the invokeRepeatFinder to set
     */
    public void setInvokeRepeatFinder(boolean invokeRepeatFinder) {
        this.invokeRepeatFinder = invokeRepeatFinder;
    }

    public CsvRawImporter(int separatorIndex, int nameColumn, int codeColumn, boolean includeTail,
            boolean invokeRepeatFinder, boolean verbose) {
        this(CsvImporter.getSeparator(separatorIndex), nameColumn, codeColumn, includeTail,
                invokeRepeatFinder, verbose);
    }

    public CsvRawImporter(String separator, int nameColumn, int codeColumn, boolean includeTail,
            boolean invokeRepeatFinder, boolean verbose) {
        super(separator, nameColumn, verbose);
        this.separator = separator;
        this.nameColumn = nameColumn;
        this.codeColumn = codeColumn;
        this.invokeRepeatFinder = invokeRepeatFinder;
        this.includeTail = includeTail;
    }

    private CsvRawImporter() { // ???
        this(0, -1, 1, false, true, false);
    }

    @Override
    public void load(File file) throws IOException, ParseException {
        FileInputStream stream = new FileInputStream(file);
        try {
            load(stream, file.getPath());
        } finally {
            stream.close();
        }
    }

    @Override
    public void load(Reader reader, String origin) throws IOException {
        prepareLoad(origin);

        BufferedReader bufferedReader = new BufferedReader(reader);
        lineNo = 1;
        for (String line = bufferedReader.readLine(); line != null; line = bufferedReader.readLine()) {
            String[] chunks = line.split(separator);
            Command signal = scrutinizeRaw(chunks);
            if (signal != null) {
                addCommand(signal);
            }
            lineNo++;
        }
        setupRemoteSet();
    }

    private Command scrutinizeRaw(String[] chunks) {
        String code = gobbleString(chunks, codeColumn, null, includeTail);
        if (code == null || code.isEmpty()) {
            if (verbose)
                System.err.println("Empty code in line " + lineNo);
            return null;
        }
        String name = gobbleString(chunks, nameColumn, "-", false);
        IrSignal irSignal = null;
        try {
            irSignal = Utils.interpretString(code, IrpUtils.defaultFrequency, invokeRepeatFinder);
        } catch (IrpMasterException ex) {
            if (verbose)
                System.err.println("Error parsing code in line " + lineNo + " (" + ex.getMessage() + ")");
            return null;
        }
        return new Command(name, "Line " + lineNo + ", " + origin, irSignal, generateCcf, invokeDecodeIr);
    }

    public static Collection<Command> process(File file, String separator, int nameColumn, int codeColumn, boolean includeTail,
            boolean invokeAnalyzer, boolean invokeRepeatFinder, boolean verbose) throws IOException, ParseException {
        CsvRawImporter csvImportRaw = new CsvRawImporter(separator, nameColumn, codeColumn, includeTail,
                invokeRepeatFinder, verbose);
        csvImportRaw.load(file);
        return csvImportRaw.getCommands();
    }

    public static void main(String[] args) {
        try {
            Collection<Command> signals = process(new File(args[0]), " ", 3, 6, true, true, true, true);
            for (Command signal : signals)
                System.out.println(signal.toPrintString());
        } catch (IOException ex) {
            System.err.println(ex.getMessage());
        } catch (IrpMasterException ex) {
            System.err.println(ex.getMessage());
        } catch (ParseException ex) {
            System.err.println(ex.getMessage());
        }
    }
}
